'''MagentroPy package metadata.'''

# keep consistent with pyproject.toml
__project__ = 'MagentroPy'
__author__ = 'Soren Bear'
__email__ = 'sorenbear@berkeley.edu'
__copyright__ = f'2022-2023, {__author__}'
__license__ = 'MIT'
__version__= '0.1.6'
__repo_url__ = 'https://github.com/sorenbear/magentropy'
